#!/usr/bin/env python3
"""
Collect modules and python stubs from other projects and stores them in the all_stubs folder
The all_stubs folder should be mapped/symlinked to the micropython_stubs/stubs repo/folder

 1) get cpython core modules
 2) get micropython frozen modules for the CURRENT checked out version
 3) get Loboris frozen modules (no longer maintained)
 4) Generate/update type hint files (pyi) for all stubs.

"""
# pylint: disable= line-too-long, W1202
# Copyright (c) 2020 Jos Verlinde
# MIT license
# pylint: disable= line-too-long
import logging
import basicgit as git
import utils
from pathlib import Path

import get_cpython
import get_mpy
import get_lobo

# todo: add frozen modules for : pycopy
# import freezer_pycopy

log = logging.getLogger(__name__)

STUB_FOLDER = "./all-stubs"
MPY_PATH = "./micropython"


def get_all(mpy_path=MPY_PATH):
    "get all frozen modules for the current version of micropython"

    #
    get_cpython.get_core(stub_path=utils.stubfolder("cpython_core"), requirements="./src/reqs-cpython-mpy.txt")

    version = utils.clean_version(git.get_tag(mpy_path))

    if version:
        log.info("found micropython version : {}".format(version))
        # folder/{family}-{version}-frozen
        family = "micropython"
        stub_path = utils.stubfolder("{}-{}-frozen".format(family, utils.flat_version(version)))
        get_mpy.get_frozen(stub_path, version=version, mpy_path=mpy_path, lib_path="./micropython-lib")
    else:
        log.warning("Unable to find the micropython repo in folder : {}".format(mpy_path))

    get_lobo.get_frozen(stub_path=STUB_FOLDER + "/loboris-esp32_lobo_3_2_24" + "-frozen")

    # now generate typeshed files for all scripts
    log.info("Generate type hint files (pyi) in folder: {}".format(STUB_FOLDER))
    utils.generate_pyi_files(Path(STUB_FOLDER))


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="%(name)-10s %(levelname)-8s:%(message)s")

    get_all()



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, AnyStr, cast

from micropy.exceptions import PyDeviceConnectionError
from micropy.pyd.abc import (
    DevicePath,
    HostPath,
    MessageConsumer,
    MessageHandler,
    MetaPyDeviceBackend,
    PyDeviceConsumer,
)

if TYPE_CHECKING:
    from typing import type_check_only  # pragma: no cover

    @type_check_only  # pragma: no cover
    class RShell:
        ASCII_XFER: bool
        QUIET: bool

        def connect(self, port: str):
            ...


try:
    import rshell.main as rsh  # type: ignore
    from rshell.pyboard import Pyboard, PyboardError  # type: ignore
except (
    ImportError,
    ModuleNotFoundError,
):  # pragma: no cover
    rsh = object()  # type: ignore
    if TYPE_CHECKING:
        rsh: RShell = cast(RShell, object())  # type: ignore
    PyboardError = RuntimeError
    Pyboard = object()


class RShellConsumer:
    consumer: MessageHandler

    def __init__(self, child_consumer: MessageHandler):
        self._outline: list[str] = []
        self.consumer = child_consumer

    def _output(self, data: str):
        """Yields everything up to a newline.

        Args:
            data (str): Anything to yield before newline

        """
        if data == "\n":
            line = "".join(self._outline)
            self._outline = []
            yield line
        self._outline.append(data)

    def on_message(self, char: bytes):
        """Pyboard data consumer.

        When a full line of output is detected,
        it is formatted then logged to stdout
        and log file.

        Args:
            char (byte): Byte from PyBoard

        Returns:
            str: Converted char

        """
        _char = char.decode("utf-8")
        line = next(self._output(_char), None)
        if line:
            self.consumer(line)
        return char


class RShellPyDeviceBackend(MetaPyDeviceBackend):
    _connected: bool = False
    _verbose: bool = False
    _rsh: rsh
    _pydevice: Pyboard

    _dev_port: str
    _repl_active: bool = False

    @property
    def _pyb_root(self) -> str:
        """pyboard root dirname."""
        if self.connected:
            dev = rsh.find_serial_device_by_port(self.location)
            return getattr(dev, "name_path", "/pyboard/")
        return ""

    @property
    def connected(self) -> bool:
        return self._connected

    def resolve_path(self, path: str | DevicePath | Path) -> DevicePath:
        _path = path
        if str(path)[0] == "/":
            _path = str(path)[1:]
        pyb_path = f"{self._pyb_root}{_path}"
        return DevicePath(pyb_path)

    def establish(self, target: str) -> RShellPyDeviceBackend:
        self._rsh = rsh
        self._rsh.ASCII_XFER = False
        self._rsh.QUIET = not self._verbose
        self.location = target
        return self

    def connect(self) -> None:
        try:
            self._rsh.connect(self.location)
        except (SystemExit, Exception) as e:
            raise PyDeviceConnectionError(self.location) from e
        self._connected = True
        dev = self._rsh.find_serial_device_by_port(self.location)
        if dev is None:
            raise PyDeviceConnectionError(self.location)
        self._pydevice = dev.pyb

    def disconnect(self) -> None:
        return

    def reset(self) -> None:
        return

    def push_file(self, source_path: HostPath, target_path: DevicePath = None, **_) -> None:
        """Copies file to pyboard.

        Args:
            source_path (str): path to file
            target_path (str, optional): dest on pyboard. Defaults to None.
                If None, file is copied to pyboard root.

        Returns:
            str: path to dest on pyboard

        """
        src_path = Path(source_path).resolve()
        _dest = target_path or src_path.name
        dest = self.resolve_path(_dest)
        self._rsh.cp(str(src_path), str(dest))

    def pull_file(self, source_path: DevicePath, target_path: HostPath, **kwargs) -> None:
        host_dest = Path(target_path).resolve()
        device_src = self.resolve_path(source_path)
        self._rsh.cp(str(device_src), str(host_dest))

    def list_dir(self, path: DevicePath) -> list[DevicePath]:
        """List directory on pyboard.

        Args:
            path: path to directory

        """
        dir_path = self.resolve_path(path)
        tree = self._rsh.auto(rsh.listdir, str(dir_path))
        return tree

    def copy_dir(
        self, source_path: HostPath | DevicePath, target_path: HostPath | DevicePath, **rsync
    ):
        """Copy directory from pyboard to machine.

        Args:
            source_path: path to directory
            target_path: destination to copy to
            rsync (dict, optional): additonal args to pass to rsync call.
                Defaults to {}

        """
        dir_path = self.resolve_path(source_path)
        dest_path = Path(str(target_path))
        rsync_args = {
            "recursed": True,
            "mirror": False,
            "dry_run": False,
            "print_func": lambda *args: None,
            "sync_hidden": False,
        }
        self._rsh.rsync(dir_path, str(dest_path), **(rsync_args | rsync))
        return dest_path

    @contextmanager
    def repl(self):
        """Pyboard raw repl context manager."""
        if self._repl_active:
            yield self._pydevice
        else:
            self._pydevice.enter_raw_repl()
            self._repl_active = True
            try:
                yield self._pydevice
            finally:
                self._pydevice.exit_raw_repl()
                self._repl_active = False

    def eval(self, command: str, *, consumer: MessageConsumer = None):
        """Execute bytes on pyboard."""
        _handler = None if consumer is None else RShellConsumer(consumer.on_message).on_message
        ret, ret_err = self._pydevice.exec_raw(command, data_consumer=_handler)
        if ret_err:
            raise PyboardError("exception", ret, ret_err)
        return ret

    def eval_script(
        self,
        contents: AnyStr,
        target_path: DevicePath | None = None,
        *,
        consumer: PyDeviceConsumer = None,
    ):
        _contents: str | bytes = contents
        if isinstance(_contents, bytes):
            _contents = _contents.decode()
        with self.repl():
            try:
                out_bytes = self.eval(_contents, consumer=consumer)
            except PyboardError as e:
                raise Exception(str(e))
            out = out_bytes.decode("utf-8")
            return out



""" 
    Read the Micropython library documentation files and use them to build stubs that can be used for static typechecking 
    using a custom-built parser to read and process the micropython RST files
    - generates:
        - modules 
            - docstrings

        - function definitions 
            - function parameters based on documentation
            - docstrings

        classes
            - docstrings
            - __init__ method
            - parameters based on documentation for class 
            - methods
                - parameters based on documentation for the method
                - docstrings

        - exceptions

    - tries to determine the return type by parsing the docstring.
        - Imperative verbs used in docstrings have a strong correlation to return -> None
        - recognizes documented Generators, Iterators, Callable
        - Coroutines are identified based tag "This is a Coroutine". Then if the return type was Foo, it will be transformed to : Coroutine[Foo]
        - a static Lookup list is used for a few methods/functions for which the return type cannot be determined from the docstring. 
        - add NoReturn to a few functions that never return ( stop / deepsleep / reset )
        - if no type can be detected the type `Any` is used

    The generated stub files are formatted using `black` and checked for validity using `pyright`
    Note: black on python 3.7 does not like some function defs 
    `def sizeof(struct, layout_type=NATIVE, /) -> int:` 

    - ordering of inter-dependent classes in the same module   
    
    - Literals / constants
        - documentation contains repeated vars with the same indentation
        - Module level:
        .. code-block:: 

            .. data:: IPPROTO_UDP
                     IPPROTO_TCP

        - class level: 
        .. code-block:: 
        
            .. data:: Pin.IRQ_FALLING
                    Pin.IRQ_RISING
                    Pin.IRQ_LOW_LEVEL
                    Pin.IRQ_HIGH_LEVEL

                    Selects the IRQ trigger type.

        - literals documented using a wildcard are added as comments only 

    - add GLUE imports to allow specific modules to import specific others. 

    - repeats of definitions in the rst file for similar functions or literals
        - CONSTANTS ( module and Class level )
        - functions
        - methods

    - Child/ Parent classes
        are added based on a (manual) lookup table CHILD_PARENT_CLASS



Not yet implemented 
-------------------

    - manual tweaks for uasync 
        https://docs.python.org/3/library/typing.html#asynchronous-programming


    # correct warnings for 'Unsupported escape sequence in string literal'


"""

import sys
import json
import re
import subprocess
import basicgit as git
from typing import List, Tuple
from pathlib import Path
from utils import flat_version

from rst import (
    return_type_from_context,
    TYPING_IMPORT,
    MODULE_GLUE,
    PARAM_FIXES,
    CHILD_PARENT_CLASS,
    ModuleSourceDict,
    ClassSourceDict,
    FunctionSourceDict,
)

NEW_OUTPUT = True
#: self.gather_docs = True
SEPERATOR = "::"


def _color(c, *args) -> str:
    s = str(*args)
    return f"\033[{c}m {s}\033[00m"


def _green(*args) -> str:
    return _color(92, *args)


def _red(*args) -> str:
    return _color(91, *args)


class RSTReader:
    __debug = False  # a
    verbose = True
    gather_docs = False  # used only during Development
    target = ".py"  # py/pyi

    def __init__(self, v_tag="v1.xx"):
        self.line_no: int = 0  # current Linenumber used during parsing.
        self.filename = ""

        self.current_module = ""
        self.current_class = ""
        self.current_function = ""  # function & method

        # input buffer
        self.rst_text: List[str] = []
        self.max_line = 0

        # Output buffer
        self.output: List[str] = []
        self.output_dict: ModuleSourceDict = ModuleSourceDict("")
        self.output_dict.add_import(TYPING_IMPORT)

        self.source_tag = v_tag

        # development aids only
        self.return_info: List[Tuple] = []
        self.last_line = ""

    @property
    def line(self) -> str:
        "get the current line from input, also stores this as last_line to allow for inspection and dumping the json file"
        if self.line_no >= 0 and self.line_no <= self.max_line:
            self.last_line = self.rst_text[self.line_no]
        else:
            self.last_line = ""
        return self.last_line

    @property
    def module_names(self) -> List[str]:
        "list of possible module names [uname , name] (longest first)"
        namelist: List[str] = []
        if self.current_module == "":
            return namelist
        # deal with module names "esp and esp.socket"
        if "." in self.current_module:
            names = [self.current_module, self.current_module.split(".")[0]]
        else:
            names = [self.current_module]
        # process
        for c_mod in names:
            if self.current_module[0] != "u":
                namelist += [f"u{c_mod}", c_mod]
            else:
                namelist += [c_mod, c_mod[1:]]
        return namelist

    def strip_prefixes(self, name: str, strip_mod: bool = True, strip_class: bool = False):
        "Remove the modulename. and or the classname. from the begining of a name"
        if strip_mod:
            prefixes = self.module_names
        else:
            prefixes = []
        if strip_class and self.current_class != "":
            prefixes += [self.current_class]
        for prefix in prefixes:
            if len(prefix) > 1 and prefix + "." in name:
                name = name.replace(prefix + ".", "")
        return name

    def log(self, *arg):
        if self.verbose:
            print(*arg)

    def leave_class(self):
        if self.current_class != "":
            self.current_class = ""

    def read_file(self, filename: Path):
        print(f"::group::[Reading RST] {filename}")
        # ingore Unicode decoding issues
        with open(filename, errors="ignore", encoding="utf8") as file:
            self.rst_text = file.readlines()
        self.filename = filename
        self.max_line = len(self.rst_text) - 1
        self.current_module = filename.stem  # just to be sure

    def prepare_output(self):
        "clean up some trailing spaces and commas"
        if NEW_OUTPUT:
            lines = str(self.output_dict).splitlines(keepends=True)
            self.output = lines
        for i in range(0, len(self.output)):
            if "(self, ) ->" in self.output[i]:
                self.output[i] = self.output[i].replace("(self, ) ->", "(self) ->")

    def write_file(self, filename: Path) -> bool:
        self.prepare_output()
        try:
            print(f" - Writing to: {filename}")
            with open(filename, mode="w", encoding="utf8") as file:
                file.writelines(self.output)
        except OSError as e:
            print(e)
            return False
        if self.gather_docs:
            print(f" - Writing to: {filename}")
            with open(filename.with_suffix(".json"), mode="w", encoding="utf8") as file:
                json.dump(self.return_info, file, ensure_ascii=False, indent=4)
            self.return_info = []

        return True

    def parse_docstring(self) -> List[str]:
        """Read a textblock that will be used as a docstring, or used to process a toc tree
        The textblock is terminated at the following RST line structures/tags

            .. <anchor>
            -- Heading
            == Heading
            ~~ Heading

        The blank lines at the start and end are removed to limit the space the docstring takes up.
        """
        if self.line_no >= len(self.rst_text):
            raise IndexError
        block: List[str] = []
        self.line_no += 1  # advance over current line
        try:
            while (
                self.line_no < len(self.rst_text)
                and not self.rst_text[self.line_no]
                .lstrip()
                .startswith("..")  # stop at anchor ( however .. note: could be considered to be added)
                and not self.rst_text[min(self.line_no + 1, self.max_line - 1)].startswith("--")  # Heading --
                and not self.rst_text[min(self.line_no + 1, self.max_line - 1)].startswith("==")  # Heading ==
                and not self.rst_text[min(self.line_no + 1, self.max_line - 1)].startswith("~~")  # Heading ~~
            ):
                line = self.rst_text[self.line_no]
                block.append(line.rstrip())
                self.line_no += 1  # advance line
        except IndexError:
            pass
        # if a Quoted Literal Block , then remove the first character of each line
        # https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#quoted-literal-blocks
        if len(block) > 0 and len(block[0]) > 0 and block[0][0] != " ":
            q_char = block[0][0]
            if all([l.startswith(q_char) for l in block]):
                # all lines start with the same character, so skip that character
                block = [l[1:] for l in block]

        # remove empty lines at beginning/end of block
        if len(block) and len(block[0]) == 0:
            block = block[1:]
        if len(block) and len(block[-1]) == 0:
            block = block[:-1]

        # prettify
        if len(block):
            # Clean up Synopsis
            if ":synopsis:" in block[0]:
                block[0] = re.sub(
                    r"\s+:synopsis:\s+(?P<syn>[\w|\s]*)",
                    r"\g<syn>",
                    block[0],
                )
        # add clickable hyperlinks to CPython docpages
        for i in range(0, len(block)):
            # hyperlink to Cpython doc pages
            # https://regex101.com/r/5RN8rj/1
            # Optionally link to python 3.4 / 3.5 documentation
            block[i] = re.sub(
                r"(\s*\|see_cpython_module\|\s+:mod:`python:(?P<mod>[\w|\s]*)`)[.]?",
                r"\g<1> https://docs.python.org/3/library/\g<mod>.html .",
                block[i],
            )
            # RST hyperlink format is not clickable in v
            # https://regex101.com/r/5RN8rj/1
            block[i] = re.sub(
                r"(.*)(?P<url><https://docs\.python\.org/.*>)(`_)",
                r"\g<1>`\g<url>",
                block[i],
            )

        #
        return block

    def fix_parameters(self, params: str):
        "change documentation parameter notation to a supported format that works for linting"
        params = params.strip()
        if not params.endswith(")"):
            # remove all after the closing bracket
            params = params[0 : params.rfind(")") + 1]

        ## Deal with SQUARE brackets first ( Documentation meaning := [optional])

        # multiple optionals
        # # .. method:: Servo.angle([angle, time=0])
        if "[angle, time=0]" in params:
            params = params.replace("[angle, time=0]", "[angle], time=0")
        # .. method:: Servo.speed([speed, time=0])
        elif "[speed, time=0]" in params:
            params = params.replace("[speed, time=0]", "[speed], time=0")

        # change [x] --> x:Optional[Any]
        params = params.replace("[", "")
        params = params.replace("]]", "")  # Q&D Hack-complex nesting
        params = params.replace("]", ": Optional[Any]")

        # deal with overloads for Flash and Partition .readblock/writeblocks
        params = params.replace("block_num, buf, offset", "block_num, buf, offset: Optional[int]")

        # Remove modulename. and Classname. from class constant
        params = self.strip_prefixes(params, strip_mod=True, strip_class=True)

        for fix in PARAM_FIXES:
            if fix[0] in params:
                params = params.replace(fix[0], fix[1])

        # formatting
        # fixme: ... not allowed in .py
        if self.target == ".py":
            params = params.replace("*, ...", "*args")
            params = params.replace("...", "*args")

        return params

    def create_update_class(self, name: str, params: str, docstr: List[str]):
        # a bit of a hack: assume no classes in classes  or functions in function
        self.leave_class()
        # TODO: Add / update information to existing class definition
        full_name = self.output_dict.find(f"class {name}")
        if full_name:
            print(_red(f"TODO: UPDATE EXISTING CLASS : {name}"))
            class_def = self.output_dict[full_name]
        else:
            # TODO: add the parent class(es) to the formal documentation
            parent = ""
            if name in CHILD_PARENT_CLASS.keys():
                parent = CHILD_PARENT_CLASS[name]

            class_def = ClassSourceDict(
                f"class {name}({parent}):",
                docstr=docstr,
            )
        if len(params) > 0:
            method = FunctionSourceDict(
                name="__init__",
                indent=class_def._indent + 4,
                definition=[f"def __init__(self, {params} -> None:"],
                docstr=[],  # todo: check if twice is needed
            )
            class_def += method
        # Append class to output
        self.output_dict += class_def
        self.current_class = name

    def parse_toc(self):
        "process table of content with additional rst files, and add / include them in the current module"
        self.log(f"# {self.line.rstrip()}")
        self.line_no += 1  # skip one line
        toctree = self.parse_docstring()
        # cleanup toctree
        toctree = [x.strip() for x in toctree if f"{self.current_module}." in x]
        # Now parse all files mentioned in the toc
        for file in toctree:
            self.read_file(Path("micropython/docs/library") / file.strip())
            self.parse()
        # reset this file to done
        self.rst_text = []
        self.line_no = 1

    def parse_module(self):
        "parse a module tag and set the module's docstring"
        self.log(f"# {self.line.rstrip()}")
        module_name = self.line.split(SEPERATOR)[-1].strip()

        self.current_module = module_name
        self.current_function = self.current_class = ""
        # get module docstring
        docstr = self.parse_docstring()

        if len(docstr) > 0:
            # Add link to online documentation
            # https://docs.micropython.org/en/v1.17/library/array.html
            if "nightly" in self.source_tag:
                version = "latest"
            else:
                version = self.source_tag.replace("_", ".")
            docstr[0] = docstr[0] + f". See: https://docs.micropython.org/en/{version}/library/{module_name}.html"

        self.output_dict.name = module_name
        self.output_dict.add_comment(f"# source version: {self.source_tag}")
        self.output_dict.add_comment(f"# origin module:: {self.filename}")
        self.output_dict.add_docstr(docstr)
        # Add additional imports to allow one module te refer to another
        if module_name in MODULE_GLUE.keys():
            self.output_dict.add_import(MODULE_GLUE[module_name])

    def parse_current_module(self):
        self.log(f"# {self.line.rstrip()}")
        module_name = self.line.split(SEPERATOR)[-1].strip()
        mod_comment = f"# + module: {self.current_module}.rst"
        self.current_module = module_name
        self.current_function = self.current_class = ""
        self.log(mod_comment)
        if NEW_OUTPUT:  # new output
            self.output_dict.name = module_name
            self.output_dict.add_comment(mod_comment)

        self.line_no += 1  # advance as we did not read any docstring
        # todo: read first block and do something with it ( for a submodule)

    def parse_function(self):
        self.log(f"# {self.line.rstrip()}")
        # this_function = self.line.split(SEPERATOR)[-1].strip()
        # name = this_function

        # Get one or more names
        function_names = self.parse_names(oneliner=False)
        docstr = self.parse_docstring()

        for this_function in function_names:
            # Parse return type from docstring
            ret_type = return_type_from_context(docstring=docstr, signature=this_function, module=self.current_module)

            ...
            # defaults
            name = params = ""
            try:
                name, params = this_function.split("(", maxsplit=1)
            except ValueError:
                print(_red(this_function))
            self.current_function = name
            if name in ("classmethod", "staticmethod"):
                # skip the classmethod and static method functions
                # no use to create stubs for these
                return

            # ussl documentation uses a ssl.foobar prefix
            for mod in self.module_names:
                if name.startswith(f"{mod}."):
                    # remove module name from the start of the function name
                    name = name[len(f"{mod}.") :]
            # fixup parameters
            params = self.fix_parameters(params)
            # assume no functions in classes
            self.leave_class()
            # if function name is the same as the module
            # then this is probably documenting a class ()

            # if the function name matches the module name then threat this as a class.
            if name in self.module_names:
                # 'Promote' function to class
                self.create_update_class(name, params, docstr)
            else:
                fn_def = FunctionSourceDict(
                    name=f"def {name}",
                    definition=[f"def {name}({params} -> {ret_type}:"],
                    docstr=docstr,
                )
                self.output_dict += fn_def

    def parse_class(self):
        self.log(f"# {self.line.rstrip()}")
        this_class = self.line.split(SEPERATOR)[-1].strip()  # raw
        if "(" in this_class:
            name, params = this_class.split("(", 2)
        else:
            name = this_class
            params = ""
        name = self.strip_prefixes(name)
        self.current_class = name
        self.current_function = ""

        self.log(f"# class:: {name} - {this_class}")
        # fixup parameters
        params = self.fix_parameters(params)
        docstr = self.parse_docstring()

        if any(":noindex:" in line for line in docstr):
            # if the class docstring contains ':noindex:' on any line then skip
            self.log(f"# Skip :noindex: class {name}")
        else:
            # write a class header
            self.create_update_class(name, params, docstr)

    def get_rst_hint(self):
        "parse the '.. <rst hint>:: ' from the current line"
        m = re.search(r"\.\.\s?(\w+)\s?::\s?", self.line)
        if m:
            return m.group(1)
        else:
            return ""

    def parse_method(self):
        name = ""
        this_method = ""
        # params = ")"
        ## py:staticmethod  - py:classmethod - py:decorator
        # ref: https://sphinx-tutorial.readthedocs.io/cheatsheet/
        self.log(f"# {self.line.rstrip()}")

        ## rst_hint is used to access the method decorator ( method, staticmethod, staticmethod ... )
        rst_hint = self.get_rst_hint()

        method_names = self.parse_names(oneliner=False)
        # filter out overloads with 'param=value' description as these can't be output (currently)
        method_names = [m for m in method_names if not "param=value" in m]

        docstr = self.parse_docstring()
        for this_method in method_names:
            try:
                name, params = this_method.split("(", 1)  # split methodname from params
            except ValueError:
                name = this_method
                params = ")"
            self.current_function = name
            # self.writeln(f"# method:: {name}")
            if "." in name:
                # todo deal with longer / deeper classes
                class_name = name.split(".")[0]
                # update current for out-of sequence method processing
                self.current_class = class_name
            else:
                # if nothing specified lets assume part of current class
                class_name = self.current_class
            name = name.split(".")[-1]  # Take only the last part from Pin.toggle

            # fixup optional [] parameters and other notations
            params = self.fix_parameters(params)

            # get or create the parent class
            full_name = self.output_dict.find(f"class {class_name}")
            if full_name:
                parent_class = self.output_dict[full_name]
            else:
                # not found, create and add new class to the output dict
                parent_class = ClassSourceDict(f"class {class_name}():")
                self.output_dict += parent_class

            # parse return type from docstring
            ret_type = return_type_from_context(docstring=docstr, signature=f"{class_name}.{name}", module=self.current_module)
            # methods have 4 flavours
            #   - __init__              (self,  <params>) -> None:
            #   - classmethod           (cls,   <params>) -> <ret_type>:
            #   - staticmethod          (       <params>) -> <ret_type>:
            #   - all other methods     (self,  <params>) -> <ret_type>:
            if name == "__init__":
                method = FunctionSourceDict(
                    name=f"def {name}",
                    indent=parent_class._indent + 4,
                    definition=[f"def __init__(self, {params} -> None:"],
                    docstr=docstr,
                )
                parent_class += method
            elif rst_hint == "classmethod":
                method = FunctionSourceDict(
                    decorators=["@classmethod"],
                    name=f"def {name}",
                    indent=parent_class._indent + 4,
                    definition=[f"def {name}(cls, {params} -> {ret_type}:"],
                    docstr=docstr,
                )
                parent_class += method

            elif rst_hint == "staticmethod":
                method = FunctionSourceDict(
                    decorators=["@staticmethod"],
                    name=f"def {name}",
                    indent=parent_class._indent + 4,
                    definition=[f"def {name}({params} -> {ret_type}:"],
                    docstr=docstr,
                )
                parent_class += method
            else:  # just plain method
                method = FunctionSourceDict(
                    name=f"def {name}",
                    indent=parent_class._indent + 4,
                    definition=[f"def {name}(self, {params} -> {ret_type}:"],
                    docstr=docstr,
                )
                parent_class += method

    def parse_exception(self):
        self.log(f"# {self.line.rstrip()}")
        name = self.line.split(SEPERATOR)[1].strip()
        # TODO : check name scope : Module.class.<name>
        if "." in name:
            name = name.split(".")[-1]  # Take only the last part from Pin.toggle
        except_1 = ClassSourceDict(name=f"class {name}(BaseException) : ...", docstr=[], init="")
        self.output_dict += except_1
        # no docstream read (yet) , so need to advance to next line
        self.line_no += 1
        # class Exception(BaseException): ...

    def parse_name(self, line: str = None):
        "get the constant/function/class name from a line with an identifier"
        # this_const = self.line.split(SEPERATOR)[-1].strip()
        # name = this_const
        # name = this_const.split(".")[1]  # Take only the last part from Pin.toggle

        if line:
            return line.split(SEPERATOR)[-1].strip()
        else:
            return self.line.split(SEPERATOR)[-1].strip()

    def parse_names(self, oneliner: bool = True):
        """get a list of constant/function/class names from and following a line with an identifier
        advances the linecounter

        oneliner :  treat a line with commas as multiple names (used for constants)
        """
        names: List[str] = []
        if oneliner:
            names += self.parse_name().split(",")
        else:
            names += [self.parse_name()]

        m = re.search(r"..\s?\w+\s?::\s?", self.line)
        if m:
            col = m.end()
        else:
            raise KeyError
        counter = 1
        while (
            self.line_no + counter <= self.max_line
            and self.rst_text[self.line_no + counter].startswith(" " * col)
            and not self.rst_text[self.line_no + counter][col + 1].isspace()
        ):
            if self.verbose:
                print(_green("Sequence detected"))
            names.append(self.parse_name(self.rst_text[self.line_no + counter]))
            counter += 1
        # now advance the linecounter
        self.line_no += counter - 1
        # clean up before returning
        return [n.strip() for n in names if n.strip() != "etc."]

    def parse_data(self):
        # todo: find a way to reliably add Constants at the correct level
        # Note : makestubs has no issue with this
        self.log(f"# {self.line.rstrip()}")
        # Get one or more names
        names = self.parse_names()

        # get module docstring
        docstr = self.parse_docstring()

        # deal with documentation wildcards
        for name in names:

            type = return_type_from_context(docstring=docstr, signature=name, module=self.current_module, literal=True)
            if type in ["None"]:  # None does not make sense
                type = "Any"  # perhaps default to Int ?
            name = self.strip_prefixes(name)
            self.output_dict.add_constant_smart(name=name, type=type, docstr=docstr)

    def parse(self):
        self.line_no = 0
        while self.line_no < len(self.rst_text):
            line = self.line
            rst_hint = self.get_rst_hint()
            #    self.writeln(">"+line)
            if rst_hint == "module":
                self.parse_module()
            elif rst_hint == "currentmodule":
                self.parse_current_module()
            elif rst_hint == "function":
                self.parse_function()

            elif rst_hint == "class":
                self.parse_class()
            elif rst_hint == "method" or rst_hint == "staticmethod" or rst_hint == "classmethod":
                self.parse_method()
            elif rst_hint == "exception":
                self.parse_exception()
            elif rst_hint == "data":
                self.parse_data()

            elif rst_hint == "toctree":
                self.parse_toc()
                # not this will be the end of this file processing.

            elif len(rst_hint) > 0:
                # something new / not yet parsed
                self.log(f"# {line.rstrip()}")
                if self.verbose:
                    print(_red(line.rstrip()))
                self.line_no += 1
            else:
                # NOTHING TO SEE HERE , MOVE ON
                self.line_no += 1
            #  TODO: rebuild this into the NEW_OUTPUT way of working
            #     if self.gather_docs and len(docstr) > 0:
            #         self.return_info.append(
            #             (
            #                 self.current_module,
            #                 self.current_class,
            #                 self.current_function,
            #                 self.last_line,
            #                 docstr,
            #             )
            #         )

        # now clean up
        self.prepare_output()


def generate_from_rst(rst_folder: Path, dst_folder: Path, v_tag: str, black=True, pattern: str = "*.rst") -> int:
    if not dst_folder.exists():
        dst_folder.mkdir(parents=True)
    # no index, and module.xxx.rst is included in module.py
    files = [f for f in rst_folder.glob(pattern) if f.stem != "index" and "." not in f.stem]
    for file in files:
        reader = RSTReader(v_tag)
        reader.read_file(file)
        reader.parse()
        reader.write_file((dst_folder / file.name).with_suffix(".py"))
        del reader

    if black:
        try:

            cmd = ["black", str(dst_folder)]
            if sys.version_info.major == 3 and sys.version_info.minor == 7:
                # black on python 3.7 does not like some function defs
                # def sizeof(struct, layout_type=NATIVE, /) -> int:
                cmd += ["--fast"]
            result = subprocess.run(cmd, capture_output=False, check=True, shell=True)
            if result.returncode != 0:
                raise Exception(result.stderr.decode("utf-8"))
        except subprocess.SubprocessError:
            print(_red("some of the files are not in a proper format"))

    return len(files)


if __name__ == "__main__":
    base_path = Path("micropython")
    # base_path = Path("../pycopy")
    v_tag = git.get_tag(base_path.as_posix())
    if not v_tag:
        # if we can't find a tag , bail
        raise ValueError

    rst_folder = base_path / "docs" / "library"
    dst_folder = Path("all-stubs") / base_path.stem / (flat_version(v_tag, keep_v=True) + "-docs")

    generate_from_rst(rst_folder, dst_folder, v_tag)
    # generate_from_rst(rst_folder, dst_folder, v_tag, pattern="binascii.rst")  # debug



# -*- coding: utf-8 -*-

"""Packages Module.

Allows user to address different dependency types (package, module,
path, pypi, etc.) through a single uniform api.

"""

from pathlib import Path
from typing import Any, Optional, Union

import requirements

from .package import Package
from .source_package import PackageDependencySource, VCSDependencySource
from .source_path import LocalDependencySource


def create_dependency_source(
    requirement: str, name: Optional[str] = None, **kwargs: Any
) -> Union[LocalDependencySource, PackageDependencySource, VCSDependencySource]:
    """Factory for creating a dependency source object.

    Args:
        requirement (str): Package name/path/constraints in string form.
        name (str, optional): Override package name.
            Defaults to None.

    Returns:
        Appropriate Dependency Source

    """
    req = next(requirements.parse(str(requirement)))
    if req.local_file:
        path = Path(req.path)
        name = name or path.name
        pkg = Package(name, req.specs, path=req.path)
        source = LocalDependencySource(pkg, path)
        return source
    pkg = Package(**req.__dict__)
    if pkg.vcs is not None or pkg.revision is not None:
        return VCSDependencySource(pkg, **kwargs)
    return PackageDependencySource(pkg, **kwargs)


__all__ = [
    "Package",
    "PackageDependencySource",
    "LocalDependencySource",
    "create_dependency_source",
]



"""
Helper module to create stubs for the lvgl modules.
Note that the stubs can be very large, and it may be best to directly store them on an SD card if your device supports this.
"""

try:
    import lvgl
except Exception:
    print("\n\nNOTE: The `lvgl` module could not be found on this firmware\n\n")

import sys

## Note that on MicroPython this will still cause the full stubbing to be automatically started due to the IsMicroPython()
from createstubs import Stubber, __version__, isMicroPython, logging


def main():
    "Create stubs for the lvgl modules using the lvlg version number."
    print("stubber version :", __version__)
    try:
        logging.basicConfig(level=logging.INFO)
        # logging.basicConfig(level=logging.DEBUG)
    except NameError:
        pass
    # Specify firmware name & version
    try:
        fw_id = "lvgl-{0}_{1}_{2}-{3}-{4}".format(
            lvgl.version_major(),
            lvgl.version_minor(),
            lvgl.version_patch(),
            lvgl.version_info(),
            sys.platform,
        )
    except Exception:
        fw_id = "lvgl-{0}_{1}_{2}_{3}-{4}".format(8, 1, 0, "dev", sys.platform)
    stubber = Stubber(firmware_id=fw_id, path="/sd")
    stubber.clean()

    # Just get the lvlg modules written in C
    stubber.modules = ["io", "lodepng", "rtch", "lvgl"]
    stubber.create_all_stubs()
    stubber.report()


if __name__ == "__main__" or isMicroPython():
    main()



#!/usr/bin/env python3
"""
Collect modules and python stubs from other projects and stores them in the all_stubs folder
The all_stubs folder should be mapped/symlinked to the micropython_stubs/stubs repo/folder

"""
# pylint: disable= line-too-long, W1202
# Copyright (c) 2020 Jos Verlinde
# MIT license
# pylint: disable= line-too-long
import logging
import utils
import sys
from pathlib import Path

log = logging.getLogger(__name__)

if __name__ == "__main__":
    # generate typeshed files for all scripts
    if len(sys.argv) == 2:
        stub_path = sys.argv[1]
    else:
        stub_path = utils.STUB_FOLDER
    stub_path = Path(stub_path).absolute()
    log.info("Generate type hint files (pyi) in folder: {}".format(stub_path))
    utils.generate_pyi_files(stub_path)



# -*- coding: utf-8 -*-

"""
micropy.utils.helpers
~~~~~~~~~~~~~~

This module contains generic utility helpers
used by MicropyCli
"""

import inspect
import io
import shutil
import subprocess as subproc
import sys
import tarfile
import xml.etree.ElementTree as ET
from datetime import timedelta
from pathlib import Path

import micropy
import requests
import requirements
from cachier import cachier
from packaging import version
from requests import exceptions as reqexc
from requests import utils as requtil
from tqdm import tqdm

__all__ = [
    "is_url",
    "get_url_filename",
    "ensure_existing_dir",
    "ensure_valid_url",
    "is_downloadable",
    "is_existing_dir",
    "stream_download",
    "search_xml",
    "get_package_meta",
    "extract_tarbytes",
    "iter_requirements",
    "create_dir_link",
    "is_dir_link",
    "is_update_available",
    "get_cached_data",
    "get_class_that_defined_method",
]


def is_url(url):
    """Check if provided string is a url.

    Args:
        url (str): url to check

    Returns:
        bool: True if arg url is a valid url

    """
    scheme = requtil.urlparse(str(url)).scheme
    return scheme in (
        "http",
        "https",
    )


@cachier(stale_after=timedelta(days=1))
def ensure_valid_url(url):
    """Ensure a url is valid.

    Args:
        url (str): URL to validate

    Raises:
        InvalidURL: URL is not a valid url
        ConnectionError: Failed to connect to url
        HTTPError: Reponse was not 200 <OK>

    Returns:
        str: valid url

    """
    if not is_url(url):
        raise reqexc.InvalidURL(f"{url} is not a valid url!")
    resp = requests.head(url, allow_redirects=True)
    resp.raise_for_status()
    return url


def ensure_existing_dir(path):
    """Ensure path exists and is a directory.

    If path does exist, it will be returned as
    a pathlib.PurePath object

    Args:
        path (str): path to validate and return

    Raises:
        NotADirectoryError: path does not exist
        NotADirectoryError: path is not a directory

    Returns:
        object: pathlib.PurePath object

    """
    _path = Path(path)
    path = _path.absolute()
    try:
        if not path.exists():
            raise NotADirectoryError(f"{_path} does not exist!")
        if not path.is_dir():
            raise NotADirectoryError(f"{_path} is not a directory!")
    except OSError:
        raise NotADirectoryError(f"{_path} is not a valid path!")
    return _path


def is_existing_dir(path):
    """Check if path is an existing directory.

    Args:
        path (str): path to check

    Returns:
        bool: True if path exists and is a directory

    """
    try:
        ensure_existing_dir(path)
    except NotADirectoryError:
        return False
    else:
        return True


def is_downloadable(url):
    """Checks if the url can be downloaded from.

    Args:
        url (str): url to check

    Returns:
        bool: True if contains a downloadable resource

    """
    try:
        ensure_valid_url(url)
    except Exception:
        return False
    headers = requests.head(url).headers
    content_type = headers.get("content-type").lower()
    ctype = content_type.split("/")
    if any(
        t
        in (
            "text",
            "html",
        )
        for t in ctype
    ):
        return False
    return True


def get_url_filename(url):
    """Parse filename from url.

    Args:
        url (str): url to parse

    Returns:
        str: filename of url

    """
    path = requtil.urlparse(url).path
    file_name = Path(path).name
    return file_name


def stream_download(url, **kwargs):
    """Stream download with tqdm progress bar.

    Args:
        url (str): url to file

    Returns:
        bytearray: bytearray of content

    """
    stream = requests.get(url, stream=True)
    content = bytearray()
    total_size = int(stream.headers.get("content-length", len(stream.content)))
    block_size = 32 * 1024
    bar_format = "{l_bar}{bar}| [{n_fmt}/{total_fmt} @ {rate_fmt}]"
    tqdm_kwargs = {
        "unit_scale": True,
        "unit_divisor": 1024,
        "smoothing": 0.1,
        "bar_format": bar_format,
    }
    tqdm_kwargs.update(kwargs)
    with tqdm(total=total_size, unit="B", **tqdm_kwargs) as pbar:
        for block in stream.iter_content(block_size):
            pbar.update(len(block))
            content.extend(block)
    return content


@cachier(stale_after=timedelta(days=3))
def search_xml(url, node):
    """Search xml from url by node.

    Args:
        url (str): url to xml
        node (str): node to search for

    Returns:
        [str]: matching nodes

    """
    resp = requests.get(url)
    xml = resp.content.decode("UTF-8")
    root = ET.fromstring(xml)
    root_ns = root.tag[1 : root.tag.find("}")]
    namespace = {"ns": root_ns}
    _results = root.findall(f"./*/ns:{node}", namespace)
    results = [k.text for k in _results]
    return results


def iter_requirements(path):
    """Iterate requirements from a requirements.txt file.

    Args:
        path (str): path to file

    """
    req_path = Path(path).absolute()
    with req_path.open("r") as rfile:
        for req in requirements.parse(rfile):
            yield req


def get_package_meta(name, url):
    """Retrieve package metadata from PyPi.

    Args:
        name (str): Name of package with specs.
        url (str): Url to package.

    Returns:
        dict: Dictionary of Metadata

    """

    def _iter_compare(in_val, comp_to, operator):
        for t in comp_to:
            state = eval(f"in_val {operator} t")
            if state:
                yield t

    resp = requests.get(url)
    data = resp.json()
    pkg = next(requirements.parse(name))
    releases = data["releases"]
    # Latest version
    spec_data = list(releases.items())[-1][1]
    if pkg.specs:
        spec_comp, spec_v = pkg.specs[0]
        spec_v = version.parse(spec_v)
        rel_versions = [version.parse(k) for k in releases.keys()]
        spec_key = str(next(_iter_compare(spec_v, rel_versions, spec_comp)))
        spec_data = releases[spec_key]
    # Find .tar.gz meta
    tar_meta = next((i for i in spec_data if ".tar.gz" in Path(i["url"]).name))
    return tar_meta


def extract_tarbytes(file_bytes, path):
    """Extract tarfile as bytes.

    Args:
        file_bytes (bytearray): Bytes of file to extract
        path (str): Path to extract it to

    Returns:
        path: destination path

    """
    tar_bytes_obj = io.BytesIO(file_bytes)
    with tarfile.open(fileobj=tar_bytes_obj, mode="r:gz") as tar:
        tar.extractall(path)
    return path


def create_dir_link(source, target):
    """Creates a platform appropriate directory link.

    On POSIX systems it will create a symlink.
    On Windows it will fallback on a directory junction if needed

    Args:
        source (os.Pathlike): Path to create link at.
        target (os.Pathlike): Path to link to.

    Raises:
        OSError: Symlink Creation Failed
        OSError: Symlink and Directory Junction Fallback Failed

    """
    platform = sys.platform
    source = Path(source)
    target = Path(target)
    try:
        source.symlink_to(target, target_is_directory=True)
    except OSError as e:
        # Handle non-admin/non-dev windows links
        if not platform == "win32":
            # handles exFAT disk format (links not working)
            if e.errno == 38:
                shutil.copytree(
                    str(target.absolute()), str(source.absolute()), symlinks=False, ignore=None
                )
                return
            elif e.errno == 17:  # folder exists
                return
            else:
                raise e

        # Fall back to directory junction
        cmd = ["MKLINK", "/J", str(source.absolute()), str(target.absolute())]
        exit_code = subproc.call(cmd, shell=True, stdout=subproc.PIPE, stderr=subproc.PIPE)
        if exit_code:
            raise e


def is_dir_link(path):
    """Test if path is either a symlink or directory junction.

    Args:
        path (os.Pathlike): Path to test.

    Returns:
        bool: True if path is a type of link.

    """
    platform = sys.platform
    path = Path(path)
    if path.is_symlink():
        return True
    if platform == "win32":
        # Test for Directory Junction
        resolved = str(path.resolve())
        if not str(path.absolute()) == resolved:
            return True
    return False


def is_update_available():
    """Check if micropy-cli update is available.

    Returns:
        bool: True if update available, else False.

    """
    url = f"https://pypi.org/pypi/micropy-cli/json"
    data = get_cached_data(url)
    versions = [k for k in data["releases"].keys() if "rc" not in k]
    if versions:
        latest = version.parse(versions[-1])
        cur_version = version.parse(micropy.__version__)
        if cur_version < latest:
            return str(latest)
    return False


@cachier(stale_after=timedelta(days=3))
def get_cached_data(url):
    """Wrap requests with a short cache."""
    source_data = requests.get(url).json()
    return source_data


def get_class_that_defined_method(meth):
    """Determines Class that defined a given method.

    See - https://stackoverflow.com/a/25959545

    Args:
        meth (Callable): Method to determine class from

    Returns:
        Callable: Class that defined method

    """
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__):
            if cls.__dict__.get(meth.__name__) is meth:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(
            inspect.getmodule(meth), meth.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]
        )
        if isinstance(cls, type):
            return cls
    return getattr(meth, "__objclass__", None)  # handle special descriptor objects



# -*- coding: utf-8 -*-


import shutil
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, Callable, List, Optional, Tuple, Union

from git import Repo
from micropy import utils
from micropy.exceptions import RequirementNotFound

from .package import Package
from .source import DependencySource


class PackageDependencySource(DependencySource):
    """Dependency Source for pypi packages.

    Args:
        package (Package): Package source points too.
        format_desc: Callback to format progress bar description.
            Defaults to None.

    """

    repo: str = "https://pypi.org/pypi/{name}/json"

    def __init__(self, package: Package, format_desc: Optional[Callable[..., Any]] = None):
        super().__init__(package)
        try:
            utils.ensure_valid_url(self.repo_url)
        except Exception:
            raise RequirementNotFound(f"{self.repo_url} is not a valid url!", package=self.package)
        else:
            self._meta: dict = utils.get_package_meta(str(self.package), self.repo_url)
            self.format_desc = format_desc or (lambda n: n)

    @property
    def repo_url(self) -> str:
        _url = self.repo.format(name=self.package.name)
        return _url

    @property
    def source_url(self) -> str:
        return self._meta.get("url", None)

    @property
    def file_name(self) -> str:
        return utils.get_url_filename(self.source_url)

    def fetch(self) -> bytes:
        """Fetch package contents into memory.

        Returns:
            bytes: Package archive contents.

        """
        self.log.debug(f"fetching package: {self.file_name}")
        desc = self.format_desc(self.file_name)
        content = utils.stream_download(self.source_url, desc=desc)
        return content

    def __enter__(self) -> Union[Path, List[Tuple[Path, Path]]]:
        """Prepare Pypi package for installation.

        Extracts the package into a temporary directory then
        generates stubs for type hinting.
        This helps with intellisense.

        If the dependency is a module, a list
        of tuples with the file and stub path, respectively,
        will be returned. Otherwise, the path to the package
        root will be returned.

        Returns:
            Root package path or list of files.

        """
        self.tmp_path = Path(mkdtemp())
        with self.handle_cleanup():
            path = utils.extract_tarbytes(self.fetch(), self.tmp_path)
            stubs = self.generate_stubs(path)
            pkg_root = self.get_root(path)
        return pkg_root or stubs

    def __exit__(self, *args):
        shutil.rmtree(self.tmp_path, ignore_errors=True)
        return super().__exit__(*args)


class VCSDependencySource(DependencySource):
    """Dependency Source for vcs packages."""

    def __init__(self, package: Package, format_desc: Optional[Callable[..., Any]] = None):
        super().__init__(package)
        self.log.debug(
            f"VCS package!, {self.package.revision}@{self.package.vcs}@{self.package.full_name}"
        )
        self._repo: Optional[Repo] = None
        try:
            utils.ensure_valid_url(self.repo_url)
        except Exception:
            raise RequirementNotFound(
                f"{self.repo_url} is not a valid VCS url!", package=self.package
            )
        else:
            self.format_desc = format_desc or (lambda n: n)

    @property
    def repo_url(self) -> str:
        return self.package.uri

    @property
    def source_url(self) -> str:
        return self.package.uri

    @property
    def file_name(self) -> str:
        return self.package.name

    def fetch(self, dest_path: Path) -> Path:
        """Clones VCS repository to a given directory.

        Args:
            dest_path: Path to clone directory too.

        Returns:
            Path to clone repository.

        """
        self.log.debug(f"fetching vcs package: ${self.file_name} @ ${self.repo_url}")
        self.format_desc(self.file_name)
        self._repo = Repo.clone_from(self.repo_url, str(dest_path))
        return dest_path

    def __enter__(self) -> Union[Path, List[Tuple[Path, Path]]]:
        """Prepare VCS repository for installation.

        See PackageDependencySource.__enter__

        Returns:
            Root package path or list of files.

        """
        self.tmp_path = Path(mkdtemp())
        with self.handle_cleanup():
            path = self.fetch(self.tmp_path)
            stubs = self.generate_stubs(path)
            pkg_root = self.get_root(path)
        return pkg_root or stubs

    def __exit__(self, *args):
        shutil.rmtree(self.tmp_path, ignore_errors=True)
        return super().__exit__(*args)



"""
Create stubs for (all) modules on a MicroPython board
Copyright (c) 2019-2021 Jos Verlinde
"""
# pylint: disable= invalid-name, missing-function-docstring, import-outside-toplevel, logging-not-lazy
import sys
import gc
import logging
import uos as os
from utime import sleep_us
from ujson import dumps

__version__ = "1.4.3"
ENOENT = 2
_MAX_CLASS_LEVEL = 2  # Max class nesting
# deal with ESP32 firmware specific implementations.
try:
    from machine import resetWDT  # type: ignore  - LoBo specific function
except ImportError:
    # machine.WDT.feed()
    def resetWDT():
        pass


class Stubber:
    "Generate stubs for modules in firmware"

    def __init__(self, path: str = None, firmware_id: str = None):
        try:
            if os.uname().release == "1.13.0" and os.uname().version < "v1.13-103":
                raise NotImplementedError("MicroPython 1.13.0 cannot be stubbed")
        except AttributeError:
            pass

        self._log = logging.getLogger("stubber")
        self._report = []
        self.info = _info()
        gc.collect()
        if firmware_id:
            self._fwid = str(firmware_id).lower()
        else:
            self._fwid = "{family}-{port}-{ver}".format(**self.info).lower()
        self._start_free = gc.mem_free()  # type: ignore

        if path:
            if path.endswith("/"):
                path = path[:-1]
        else:
            path = get_root()

        self.path = "{}/stubs/{}".format(path, self.flat_fwid).replace("//", "/")
        self._log.debug(self.path)
        try:
            ensure_folder(path + "/")
        except OSError:
            self._log.error("error creating stub folder {}".format(path))
        self.problematic = [
            "upip",
            "upysh",
            "webrepl_setup",
            "http_client",
            "http_client_ssl",
            "http_server",
            "http_server_ssl",
        ]
        self.excluded = [
            "webrepl",
            "_webrepl",
            "port_diag",
            "example_sub_led.py",
            "example_pub_button.py",
        ]
        # there is no option to discover modules from micropython, list is read from an external file.
        self.modules = []

    def get_obj_attributes(self, item_instance: object):
        "extract information of the objects members and attributes"
        _result = []
        _errors = []
        self._log.debug("get attributes {} {}".format(repr(item_instance), item_instance))
        try:
            for name in dir(item_instance):
                try:
                    val = getattr(item_instance, name)
                    # name , item_repr(value) , type as text, item_instance
                    _result.append((name, repr(val), repr(type(val)), val))
                except AttributeError as e:
                    _errors.append(
                        "Couldn't get attribute '{}' from object '{}', Err: {}".format(name, item_instance, e)
                    )
        except AttributeError as e:
            _errors.append("Couldn't get attribute '{}' from object '{}', Err: {}".format(name, item_instance, e))
        # remove internal __
        _result = [i for i in _result if not (i[0].startswith("_") and i[0] != "__init__")]
        gc.collect()
        return _result, _errors

    def add_modules(self, modules: list):
        "Add additional modules to be exported"
        self.modules = sorted(set(self.modules) | set(modules))

    def create_all_stubs(self):
        "Create stubs for all configured modules"
        self._log.info("Start micropython-stubber v{} on {}".format(__version__, self._fwid))
        gc.collect()
        for module_name in self.modules:
            self.create_one_stub(module_name)
        self._log.info("Finally done")

    def create_one_stub(self, module_name):
        # use training comma to overide black formatting to avoid minify chocking on this
        if module_name.startswith("_") and module_name != "_thread":
            self._log.warning(
                "Skip module: {:<20}        : Internal ".format(module_name),
            )
            return False
        if module_name in self.problematic:
            self._log.warning("Skip module: {:<20}        : Known problematic".format(module_name))
            return False
        if module_name in self.excluded:
            self._log.warning("Skip module: {:<20}        : Excluded".format(module_name))
            return False

        file_name = "{}/{}.py".format(self.path, module_name.replace(".", "/"))
        gc.collect()
        m1 = gc.mem_free()  # type: ignore
        self._log.info("Stub module: {:<20} to file: {:<55} mem:{:>5}".format(module_name, file_name, m1))
        try:
            self.create_module_stub(module_name, file_name)
        except OSError:
            return False
        gc.collect()
        self._log.debug("Memory     : {:>20} {:>6X}".format(m1, m1 - gc.mem_free()))  # type: ignore
        return True

    def create_module_stub(self, module_name: str, file_name: str = None):
        """Create a Stub of a single python module

        Args:
        - module_name (str): name of the module to document. This module will be imported.
        - file_name (Optional[str]): the 'path/filename.py' to write to. If omitted will be created based on the module name.
        """
        if module_name.startswith("_") and module_name != "_thread":
            self._log.warning("SKIPPING internal module:{}".format(module_name))
            return

        if module_name in self.problematic:
            self._log.warning("SKIPPING problematic module:{}".format(module_name))
            return

        if file_name is None:
            file_name = self.path + "/" + module_name.replace(".", "_") + ".py"

        if "/" in module_name:
            # for nested modules
            module_name = module_name.replace("/", ".")

        # import the module (as new_module) to examine it
        new_module = None
        try:
            new_module = __import__(module_name, None, None, ("*"))
        except ImportError:
            #            failed = True
            self._log.warning("Skip module: {:<20}        : Failed to import".format(module_name))
            return

        # Start a new file
        ensure_folder(file_name)
        with open(file_name, "w") as fp:
            # todo: improve header
            s = '"""\nModule: \'{0}\' on {1}\n"""\n# MCU: {2}\n# Stubber: {3}\n'.format(
                module_name, self._fwid, self.info, __version__
            )

            fp.write(s)
            fp.write("from typing import Any\n\n")
            self.write_object_stub(fp, new_module, module_name, "")

        self._report.append({"module": module_name, "file": file_name})
        print({"module": module_name, "file": file_name})

        if not module_name in ["os", "sys", "logging", "gc"]:
            # try to unload the module unless we use it
            try:
                del new_module
            except (OSError, KeyError):  # lgtm [py/unreachable-statement]
                self._log.warning("could not del new_module")
            try:
                del sys.modules[module_name]
            except KeyError:
                self._log.debug("could not del sys.modules[{}]".format(module_name))
        gc.collect()

    def write_object_stub(self, fp, object_expr: object, obj_name: str, indent: str, in_class: int = 0):
        "Write a module/object stub to an open file. Can be called recursive."
        gc.collect()
        if object_expr in self.problematic:
            self._log.warning("SKIPPING problematic module:{}".format(object_expr))
            return

        # self._log.debug("DUMP    : {}".format(object_expr))
        items, errors = self.get_obj_attributes(object_expr)

        if errors:
            self._log.error(errors)

        for item_name, item_repr, item_type_txt, item_instance in items:
            # # name_, repr_(value) , type as text, item_instance
            # allow the scheduler to run on LoBo based FW
            if item_name in ["classmethod", "staticmethod"]:
                continue

            resetWDT()
            sleep_us(1)

            # Class expansion only on first 3 levels (bit of a hack)
            if item_type_txt == "<class 'type'>" and len(indent) <= _MAX_CLASS_LEVEL * 4:
                self._log.debug("{0}class {1}:".format(indent, item_name))
                # stub style : generic __init__ with Empty comment and pass
                s = "\n" + indent + "class " + item_name + ":\n"  #
                s += indent + "    ''\n"

                fp.write(s)
                # self._log.debug("\n" + s)

                self._log.debug("# recursion over class {0}".format(item_name))
                self.write_object_stub(
                    fp,
                    item_instance,
                    "{0}.{1}".format(obj_name, item_name),
                    indent + "    ",
                    in_class + 1,
                )
            # Class Methods and functions
            elif "method" in item_type_txt or "function" in item_type_txt or item_name == "__init__":
                self._log.debug("# def {1} function or method, type = '{0}'".format(item_type_txt, item_name))
                # module Function or class method
                # will accept any number of params
                # return type Any
                ret = "Any"
                first = ""
                # Self parameter only on class methods/functions
                if in_class > 0:
                    first = "self, "
                    # __init__ returns None
                    if item_name == "__init__":
                        ret = "None"
                # class method - add function decoration
                if "bound_method" in item_type_txt or "bound_method" in item_repr:
                    s = "{}@classmethod\n".format(indent)
                    s += "{}def {}(cls, *args) -> {}:\n".format(indent, item_name, ret)
                else:
                    s = "{}def {}({}*args) -> {}:\n".format(indent, item_name, first, ret)
                # s += indent + "    ''\n" # EMPTY DOCSTRING
                s += indent + "    ...\n\n"
                fp.write(s)
                self._log.debug("\n" + s)
            # constants of known types & values
            elif item_type_txt == "<class 'module'>":
                # Skip imported modules
                # fp.write("# import {}\n".format(item_name))
                pass

            elif item_type_txt.startswith("<class '"):

                t = item_type_txt[8:-2]
                s = ""

                if t in ["str", "int", "float", "bool", "bytearray", "bytes"]:
                    # known type: use actual value
                    s = "{0}{1} = {2} # type: {3}\n".format(indent, item_name, item_repr, t)
                elif t in ["dict", "list", "tuple"]:
                    # dict, list , tuple: use empty value
                    ev = {"dict": "{}", "list": "[]", "tuple": "()"}
                    s = "{0}{1} = {2} # type: {3}\n".format(indent, item_name, ev[t], t)
                else:
                    # something else
                    if not t in ["object", "set", "frozenset"]:
                        # Possibly default others to item_instance object ?
                        # https://docs.python.org/3/tutorial/classes.html#item_instance-objects
                        t = "Any"
                    # Requires Python 3.6 syntax, which is OK for the stubs/pyi
                    s = "{0}{1} : {2} ## {3} = {4}\n".format(indent, item_name, t, item_type_txt, item_repr)
                fp.write(s)
                self._log.debug("\n" + s)
            else:
                # keep only the name
                self._log.debug("# all other, type = '{0}'".format(item_type_txt))
                fp.write("# all other, type = '{0}'\n".format(item_type_txt))

                fp.write(indent + item_name + " # type: Any\n")

        del items
        del errors
        try:
            del item_name, item_repr, item_type_txt, item_instance  # pylint: disable=undefined-loop-variable
        except (OSError, KeyError, NameError):  # lgtm [py/unreachable-statement]
            pass

    @property
    def flat_fwid(self):
        "Turn _fwid from 'v1.2.3' into '1_2_3' to be used in filename"
        s = self._fwid
        # path name restrictions
        chars = " .()/\\:$"
        for c in chars:
            s = s.replace(c, "_")
        return s

    def clean(self, path: str = None):
        "Remove all files from the stub folder"
        if path is None:
            path = self.path
        self._log.info("Clean/remove files in folder: {}".format(path))
        try:
            items = os.listdir(path)
        except (OSError, AttributeError):  # lgtm [py/unreachable-statement]
            # os.listdir fails on unix
            return
        for fn in items:
            item = "{}/{}".format(path, fn)
            try:
                os.remove(item)
            except OSError:
                try:  # folder
                    self.clean(item)
                    os.rmdir(item)
                except OSError:
                    pass

    def report(self, filename: str = "modules.json"):
        "create json with list of exported modules"
        self._log.info("Created stubs for {} modules on board {}\nPath: {}".format(len(self._report), self._fwid, self.path))
        f_name = "{}/{}".format(self.path, filename)
        gc.collect()
        try:
            # write json by node to reduce memory requirements
            with open(f_name, "w") as f:
                f.write("{")
                f.write(dumps({"firmware": self.info})[1:-1])
                f.write(",\n")
                f.write(dumps({"stubber": {"version": __version__}})[1:-1])
                f.write(",\n")
                f.write('"modules" :[\n')
                start = True
                for n in self._report:
                    if start:
                        start = False
                    else:
                        f.write(",\n")
                    f.write(dumps(n))
                f.write("\n]}")
            used = self._start_free - gc.mem_free()  # type: ignore
            self._log.info("Memory used: {0} Kb".format(used // 1024))
        except OSError:
            self._log.error("Failed to create the report.")


def ensure_folder(path: str):
    "Create nested folders if needed"
    i = start = 0
    while i != -1:
        i = path.find("/", start)
        if i != -1:
            if i == 0:
                p = path[0]
            else:
                p = path[0:i]
            # p = partial folder
            try:
                _ = os.stat(p)
            except OSError as e:
                # folder does not exist
                if e.args[0] == ENOENT:
                    try:
                        os.mkdir(p)
                    except OSError as e2:
                        # self._log.error("failed to create folder {}".format(p))
                        raise e2
                else:
                    # self._log.error("failed to create folder {}".format(p))
                    raise e
        # next level deep
        start = i + 1


def _info():
    "collect base information on this runtime"
    _n = sys.implementation.name  # type: ignore
    _p = sys.platform
    info = {
        "name": _n,  # - micropython
        "release": "0.0.0",  # mpy semver from sys.implementation or os.uname()release
        "version": "0.0.0",  # major.minor.0
        "build": "",  # parsed from version
        "sysname": "unknown",  # esp32
        "nodename": "unknown",  # ! not on all builds
        "machine": "unknown",  # ! not on all builds
        "family": _n,  # fw families, micropython , pycopy , lobo , pycom
        "platform": _p,  # port: esp32 / win32 / linux
        "port": _p,  # port: esp32 / win32 / linux
        "ver": "",  # short version
    }
    try:
        info["release"] = ".".join([str(n) for n in sys.implementation.version])
        info["version"] = info["release"]
        info["name"] = sys.implementation.name
        info["mpy"] = sys.implementation.mpy  # type: ignore
    except AttributeError:
        pass

    if sys.platform not in ("unix", "win32"):
        try:
            u = os.uname()
            info["sysname"] = u.sysname
            info["nodename"] = u.nodename
            info["release"] = u.release
            info["machine"] = u.machine
            # parse micropython build info
            if " on " in u.version:
                s = u.version.split(" on ")[0]
                if info["sysname"] == "esp8266":
                    # esp8266 has no usable info on the release
                    if "-" in s:
                        v = s.split("-")[0]
                    else:
                        v = s
                    info["version"] = info["release"] = v.lstrip("v")
                try:
                    info["build"] = s.split("-")[1]
                except IndexError:
                    pass
        except (IndexError, AttributeError, TypeError):
            pass

    try:  # families
        from pycopy import const  # type: ignore

        info["family"] = "pycopy"
        del const
    except (ImportError, KeyError):
        pass
    if info["platform"] == "esp32_LoBo":
        info["family"] = "loboris"
        info["port"] = "esp32"
    elif info["sysname"] == "ev3":
        # ev3 pybricks
        info["family"] = "ev3-pybricks"
        info["release"] = "1.0.0"
        try:
            # Version 2.0 introduces the EV3Brick() class.
            from pybricks.hubs import EV3Brick  # type: ignore

            info["release"] = "2.0.0"
        except ImportError:
            pass

    # version info
    if info["release"]:
        info["ver"] = "v" + info["release"].lstrip("v")
    if info["family"] != "loboris":
        if info["release"] and info["release"] >= "1.10.0" and info["release"].endswith(".0"):
            # drop the .0 for newer releases
            info["ver"] = info["release"][:-2]
        else:
            info["ver"] = info["release"]
        # add the build nr, but avoid a git commit-id
        if info["build"] != "" and len(info["build"]) < 4:
            info["ver"] += "-" + info["build"]
    # spell-checker: disable
    if "mpy" in info:  # mpy on some v1.11+ builds
        sys_mpy = int(info["mpy"])
        arch = [
            None,
            "x86",
            "x64",
            "armv6",
            "armv6m",
            "armv7m",
            "armv7em",
            "armv7emsp",
            "armv7emdp",
            "xtensa",
            "xtensawin",
        ][sys_mpy >> 10]
        if arch:
            info["arch"] = arch
    return info
    # spell-checker: enable


def get_root() -> str:
    "Determine the root folder of the device"
    r = "/flash"
    try:
        _ = os.stat(r)
    except OSError as e:
        if e.args[0] == ENOENT:
            try:
                r = os.getcwd()
            except (OSError, AttributeError):
                # unix port
                r = "."
        else:
            r = "/"
    return r


def show_help():
    print("-p, --path   path to store the stubs in, defaults to '.'")
    sys.exit(1)


def read_path() -> str:
    "get --path from cmdline. [unix/win]"
    path = ""
    if len(sys.argv) == 3:
        cmd = (sys.argv[1]).lower()
        if cmd in ("--path", "-p"):
            path = sys.argv[2]
        else:
            show_help()
    elif len(sys.argv) == 2:
        show_help()
    return path


def isMicroPython() -> bool:
    "runtime test to determine full or micropython"
    # pylint: disable=unused-variable,eval-used
    try:
        # either test should fail on micropython
        # a) https://docs.micropython.org/en/latest/genrst/syntax.html#spaces
        # Micropython : SyntaxError
        # a = eval("1and 0")  # lgtm [py/unused-local-variable]
        # Eval blocks some minification aspects

        # b) https://docs.micropython.org/en/latest/genrst/builtin_types.html#bytes-with-keywords-not-implemented
        # Micropython: NotImplementedError
        b = bytes("abc", encoding="utf8")  # lgtm [py/unused-local-variable]

        # c) https://docs.micropython.org/en/latest/genrst/core_language.html#function-objects-do-not-have-the-module-attribute
        # Micropython: AttributeError
        c = isMicroPython.__module__  # lgtm [py/unused-local-variable]
        return False
    except (NotImplementedError, AttributeError):
        return True


def main():

    stubber = Stubber(path=read_path())
    # stubber = Stubber(path="/sd")
    # Option: Specify a firmware name & version
    # stubber = Stubber(firmware_id='HoverBot v1.2.1')
    stubber.clean()
    # there is no option to discover modules from micropython, need to hardcode
    # below contains combined modules from  Micropython ESP8622, ESP32, Loboris, Pycom and ulab , lvgl
    # spell-checker: disable
    # modules to stub : 131
    stubber.modules = [
        "_onewire",
        "_thread",
        "_uasyncio",
        "ak8963",
        "apa102",
        "apa106",
        "array",
        "binascii",
        "btree",
        "builtins",
        "cmath",
        "collections",
        "crypto",
        "curl",
        "dht",
        "display",
        "display_driver_utils",
        "ds18x20",
        "errno",
        "esp",
        "esp32",
        "espidf",
        "flashbdev",
        "framebuf",
        "freesans20",
        "fs_driver",
        "functools",
        "gc",
        "gsm",
        "hashlib",
        "heapq",
        "ili9XXX",
        "imagetools",
        "inisetup",
        "io",
        "json",
        "lcd160cr",
        "lodepng",
        "logging",
        "lv_colors",
        "lv_utils",
        "lvgl",
        "lwip",
        "machine",
        "math",
        "microWebSocket",
        "microWebSrv",
        "microWebTemplate",
        "micropython",
        "mpu6500",
        "mpu9250",
        "neopixel",
        "network",
        "ntptime",
        "onewire",
        "os",
        "pyb",
        "pycom",
        "pye",
        "queue",
        "random",
        "re",
        "requests",
        "rtch",
        "select",
        "socket",
        "ssd1306",
        "ssh",
        "ssl",
        "stm",
        "struct",
        "sys",
        "time",
        "tpcalib",
        "uarray",
        "uasyncio/__init__",
        "uasyncio/core",
        "uasyncio/event",
        "uasyncio/funcs",
        "uasyncio/lock",
        "uasyncio/stream",
        "ubinascii",
        "ubluetooth",
        "ucollections",
        "ucrypto",
        "ucryptolib",
        "uctypes",
        "uerrno",
        "uftpd",
        "uhashlib",
        "uheapq",
        "uio",
        "ujson",
        "ulab",
        "ulab/approx",
        "ulab/compare",
        "ulab/fft",
        "ulab/filter",
        "ulab/linalg",
        "ulab/numerical",
        "ulab/poly",
        "ulab/user",
        "ulab/vector",
        "umachine",
        "umqtt/robust",
        "umqtt/simple",
        "uos",
        "upip",
        "upip_utarfile",
        "uqueue",
        "urandom",
        "ure",
        "urequests",
        "urllib/urequest",
        "uselect",
        "usocket",
        "ussl",
        "ustruct",
        "usys",
        "utelnetserver",
        "utime",
        "utimeq",
        "uwebsocket",
        "uzlib",
        "websocket",
        "websocket_helper",
        "writer",
        "xpt2046",
        "ymodem",
        "zlib",
    ]  # spell-checker: enable

    gc.collect()

    stubber.create_all_stubs()
    stubber.report()


if __name__ == "__main__" or isMicroPython():
    try:
        _log = logging.getLogger("stubber")
        logging.basicConfig(level=logging.INFO)
        # logging.basicConfig(level=logging.DEBUG)
    except NameError:
        pass
    main()



# -*- coding: utf-8 -*-

import json
from pathlib import Path

from boltons.fileutils import AtomicSaver

from .config_source import ConfigSource


class JSONConfigSource(ConfigSource):
    """JSON Config File Source.

    Args:
        path (Path): Path to save config too.

    """

    def __init__(self, path: Path):
        super().__init__()
        self._file_path: Path = path

    @property
    def file_path(self) -> Path:
        """Path to config file."""
        return self._file_path

    @file_path.setter
    def file_path(self, value: Path) -> Path:
        """Set config file path.

        Args:
            value (Path): New path to config file

        Returns:
            Path: Path to config file

        """
        self._file_path = value
        return self._file_path

    @property
    def exists(self) -> bool:
        return self.file_path.exists()

    def process(self) -> dict:
        """Load config from JSON file.

        Returns:
            dict: config in file

        """
        content = self.file_path.read_text()
        if not content:
            return {}
        config = json.loads(content)
        return config

    def prepare(self):
        if not self.file_path.exists():
            self.log.debug(f"creating new config file: {self.file_path}")
            self.file_path.parent.mkdir(parents=True, exist_ok=True)
            self.file_path.touch()

    def save(self, content: dict) -> Path:
        """Save current config.

        Args:
            content (dict): content to write to file.

        Returns:
            Path: path to config file.

        """
        config = json.dumps(content, indent=4, separators=(",", ": "))
        with AtomicSaver((str(self.file_path)), text_mode=True) as file:
            file.write(config)
        return self.file_path



"""
Download or update the micropyton compatibility modules from pycopy and stores them in the all_stubs folder
The all_stubs folder should be mapped/symlinked to the micropython_stubs/stubs repo/folder
"""

import os
import glob
import shutil
import subprocess
import logging
import json

import utils
from version import VERSION

log = logging.getLogger(__name__)

family = "common"


def get_core(requirements, stub_path=None):
    "Download MicroPython compatibility modules"
    if not stub_path:
        stub_path = "./all-stubs/cpython-core"

    # use pip to dowload requirements file to build folder in one go
    #   pip install --no-compile --no-cache-dir --target ./scratch/test --upgrade -r ./src/micropython.txt
    build_path = os.path.abspath("./build")
    os.makedirs(stub_path, exist_ok=True)
    os.makedirs(build_path, exist_ok=True)
    mod_manifest = None
    try:
        subprocess.run(
            [
                "pip",
                "install",
                "--target",
                build_path,
                "-r",
                requirements,
                "--no-cache-dir",
                "--no-compile",
                "--upgrade",
                "--no-binary=:all:",
            ],
            capture_output=False,
            check=True,
        )
        # build modules.json
        mod_manifest = utils.manifest(machine=family, version=VERSION)
        # copy *.py files in build folder to stub_path
        for filename in glob.glob(os.path.join(build_path, "*.py")):
            log.info("pipped : {}".format(filename))
            f_name, f_ext = os.path.splitext(os.path.basename(filename))  # pylint: disable=unused-variable
            mod_manifest["modules"].append({"file": os.path.basename(filename), "module": f_name})
            try:
                shutil.copy2(filename, stub_path)
            except OSError as err:
                log.exception(err)
    except OSError as err:
        log.error(
            "An error occurred while trying to run pip to download the MicroPython compatibility modules from PyPi: {}".format(
                err
            )
        )
    finally:
        # remove build folder
        shutil.rmtree(build_path, ignore_errors=True)
        if mod_manifest:
            # write the the module manifest for the cpython core modules
            with open(stub_path + "/modules.json", "w") as outfile:
                json.dump(mod_manifest, outfile, indent=4, sort_keys=True)


if __name__ == "__main__":
    # just run a quick test
    logging.basicConfig(format="%(levelname)-8s:%(message)s", level=logging.INFO)
    get_core(requirements="./src/reqs-cpython-mpy.txt", stub_path="./scratch/cpython_common")
